(:
Copyright (C) 2011 Adam Retter <adam.retter@googlemail.com>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
:)
xquery version "1.0";

import module namespace http = "http://expath.org/ns/http-client";
import module namespace request = "http://exist-db.org/xquery/request";
import module namespace response = "http://exist-db.org/xquery/response";
import module namespace util = "http://exist-db.org/xquery/util";

import module namespace semantic = "http://www.adamretter.org.uk/semantic" at "semantic.xqm";

declare namespace xh = "http://www.w3.org/1999/xhtml";

declare variable $local:processor-qs-param-prefix := "_enr_";
declare variable $local:processor-qs-param-url := fn:concat($local:processor-qs-param-prefix, "url");

declare variable $local:form-encoded := "application/x-www-form-urlencoded";
declare variable $local:http-moved-status := ("301", "302");

declare variable $local:allow-redirect-post-to-get := true(); (: required for some naughty websites like devon.gov.uk :)

declare variable $local:processor := xs:anyURI("http://localhost:8080/exist/rest/db/enrichment/process.xql");
declare variable $local:ignore-elements :=  ("img", "link", "script", "base", "input", "meta");
declare variable $local:rewrite-attributes := ("src", "href", "action");

declare function local:add-url-scheme-if-missing($url as xs:string) as xs:string {
    if(local:has-scheme($url)) then
        $url
    else
        (: default: assume http :)
        fn:concat("http://", $url)
};

declare function local:has-scheme($url as xs:string) as xs:boolean {
    fn:matches($url, ".*://")
};

declare function local:rewrite-url($base-uri as xs:anyURI, $url as xs:string) as xs:string {
    
    let $ext-url := 
        if(local:has-scheme($url) and fn:not(fn:starts-with($url, $base-uri)))then
            $url
        else 
            fn:concat($base-uri, fn:replace($url, $base-uri, ""))
    return
        fn:concat($local:processor, "?", $local:processor-qs-param-url, "=", fn:encode-for-uri($ext-url)) 
};

declare function local:rewrite-urls($base-uri as xs:anyURI, $response-content as node()) {
    typeswitch($response-content)
        
        case $doc as document-node() return
            document {
                for $n in $doc/node() return
                    local:rewrite-urls($base-uri, $n)
            }
        
        case $elem as element() return
            element {fn:node-name($elem) } {
                for $a in $elem/attribute() return
                    if(fn:not(fn:local-name($elem) = $local:ignore-elements) and fn:local-name($a) = $local:rewrite-attributes and fn:not(fn:contains($elem/@class, "enr-ignore"))) then
                        attribute {fn:node-name($a)} {
                            local:rewrite-url($base-uri, string($a))
                        }
                    else
                        $a
                ,
                
                for $n in $elem/node() return
                    local:rewrite-urls($base-uri, $n)
            }
            
        case $txt as text() return
            $txt
        
        default $x return
            $x
};

declare function local:get-base-uri($url as xs:string) as xs:anyURI {
    fn:concat(fn:replace($url, "(.*//.*)/.*", "$1"), "/") cast as xs:anyURI
};

declare function local:copy-current-request-headers($was-post as xs:boolean) as element(http:header)* {
    if(request:get-header("Cookie"))then
        <http:header name="Cookie" value="{request:get-header("Cookie")}"/>
    else()
};

declare function local:copy-current-request-headers() {
    local:copy-current-request-headers(false()) 
};

declare function local:form-encode-current-request-params() {
    fn:string-join(
        for $param-name in request:get-parameter-names()[fn:not(fn:starts-with(., $local:processor-qs-param-prefix))] return
            for $param-value in request:get-parameter($param-name, ()) return
                fn:concat(fn:encode-for-uri($param-name), "=", fn:encode-for-uri($param-value))
    , "&amp;")
};

declare function local:prepare-http-request($url as xs:string, $force-get as xs:boolean) as element(http:request) {

    element http:request {
        attribute href { $url },
        attribute follow-redirect { "false" },
        attribute http { "1.0" }, (: force http 1.0 for backwards compatibility :)
        
        if(fn:lower-case(request:get-method()) eq "post" and $force-get eq false())then
        (
            attribute method { "post" },
            
            if(fn:lower-case(request:get-header("Content-Type")) eq $local:form-encoded)then
            (
                local:copy-current-request-headers(true()),
                element http:body {
                    attribute media-type { $local:form-encoded },
                    text {
                        local:form-encode-current-request-params()
                    }
                }
            ) else
                local:copy-current-request-headers()
        ) else (
            attribute method { "get" },
            local:copy-current-request-headers()
        )
    }
};

declare function local:process-redirect($response-status as element(http:response)) {
    if(fn:lower-case(request:get-method()) = ("get", "head"))then
        local:retrieve-url($response-status/http:header[fn:lower-case(@name) eq "location"]/@value, false())
    
    else if(fn:lower-case(request:get-method()) = ("post"))then
        if($local:allow-redirect-post-to-get)then
            local:retrieve-url($response-status/http:header[fn:lower-case(@name) eq "location"]/@value, true())
        else()
    else()
};

declare function local:process-html-content($url as xs:string, $response-content) {
    
    let $annotated-content := semantic:annotate($url, $response-content) return
    
        local:rewrite-urls(local:get-base-uri($url), $annotated-content)
};

declare function local:retrieve-url($in-url as xs:string, $force-get as xs:boolean) {
    
    let $url := local:add-url-scheme-if-missing($in-url) return
    
        let $req := local:prepare-http-request($url, $force-get) return
        
            let $resp := http:send-request($req),
            $response-status := $resp[1] return
            
            if($response-status/@status = $local:http-moved-status)then
                local:process-redirect($response-status)
            else
                let $content-type := $response-status/http:header[@name eq "content-type"]/@value,
                $response-content := $resp[2] return
                 
                    (: html / xhtml :)
                    if(fn:starts-with($content-type, "text/html")
                        or fn:starts-with($content-type, "application/xhtml+xml")
                        or fn:starts-with($content-type, "application/xhtml"))
                    then
                        (
                            (: fixup for old browsers which dont support application/xhtml+xml :)
                            (: let $null := util:declare-option("exist:serialize", "media-type=text/html") return :)
                                local:process-html-content($url, $response-content)
                        )
                    
                    (: xml :)
                    else if(fn:starts-with($content-type, "text/xml")
                        or fn:starts-with($content-type, "application/xml")
                        or fn:matches($content-type, "application/(.*)\+xml"))
                    then
                        $response-content
                    
                    (: text :)
                    else if(fn:starts-with($content-type, "text/"))then
                        response:stream-binary(xs:base64Binary(util:base64-encode($response-content)), $content-type, fn:replace($in-url, ".*/", ""))
                    
                    (: application :)
                    else if(fn:starts-with($content-type, "application/"))then
                        response:stream-binary($response-content, $content-type, fn:replace($in-url, ".*/", ""))
                    
                    (: unknown :)
                    else
                        $response-content
};

let $in-url := request:get-parameter($local:processor-qs-param-url, ()) return
    if(fn:empty($in-url))then
        <error>Invalid URI provided</error>
    else
        local:retrieve-url($in-url, false())